# -*- coding: utf-8 -*-

from __future__ import print_function

from bisect import bisect_left
from lxml import etree
from pkg_resources import resource_filename

from . import conjugation_template
from . import exceptions

class ConjugationsParser:
    def __init__(self, lang='fr'):
        self.templates = []
        parser = etree.XMLParser(dtd_validation=True, encoding='utf-8')
        tree = etree.parse(
            resource_filename("verbecc",
                              "data/conjugations-{}.xml".format(lang)),
            parser)
        root = tree.getroot()
        root_tag = 'conjugation-{}'.format(lang)
        if root.tag != root_tag:
            raise exceptions.ConjugationsParserError(
                "Root XML Tag {} Not Found".format(root_tag))
        for child in root:
            if child.tag == 'template':
                self.templates.append(
                    conjugation_template.ConjugationTemplate(child))
        self.templates = sorted(self.templates, key=lambda x: x.name)
        self._keys = [template.name for template in self.templates]

    def find_template(self, name):
        """Assumes templates are already sorted by name"""
        i = bisect_left(self._keys, name)
        if i != len(self._keys) and self._keys[i] == name:
            return self.templates[i]
        raise exceptions.TemplateNotFoundError


if __name__ == "__main__":
    conj = ConjugationsParser()



# -*- coding: utf-8 -*-

from __future__ import print_function

from bisect import bisect_left

from lxml import etree

from pkg_resources import resource_filename

from . import string_utils
from . import verb
from . import exceptions

class VerbsParser:
    def __init__(self, lang='fr'):
        self.verbs = []
        parser = etree.XMLParser(encoding='utf-8')
        tree = etree.parse(resource_filename(
                           "verbecc",
                           "data/verbs-{}.xml".format(lang)),
                           parser)
        root = tree.getroot()
        root_tag = 'verbs-{}'.format(lang)
        if root.tag != root_tag:
            raise exceptions.VerbsParserError(
                "Root XML Tag {} Not Found".format(root_tag))
        for child in root:
            if child.tag == 'v':
                self.verbs.append(verb.Verb(child))
        self.verbs = sorted(self.verbs, key=lambda x: x.infinitive_no_accents)
        self._keys = [verb.infinitive_no_accents for verb in self.verbs]

    def find_verb_by_infinitive(self, infinitive):
        """Assumes verbs are already sorted by infinitive"""
        infinitive_no_accents = string_utils.strip_accents(infinitive.lower())
        i = bisect_left(self._keys, infinitive_no_accents)
        if i != len(self._keys) and self._keys[i] == infinitive_no_accents:
            return self.verbs[i]
        raise exceptions.VerbNotFoundError

    def get_verbs_that_start_with(self, pre, max_results=10):
        ret = []
        pre_no_accents = string_utils.strip_accents(pre.lower())
        for verb in self.verbs:
            if verb.infinitive_no_accents.startswith(pre_no_accents):
                ret.append(verb.infinitive)
                if len(ret) >= max_results:
                    break
        return ret



# -*- coding: utf-8 -*-

from verbecc import inflector

class InflectorEs(inflector.Inflector):
    def __init__(self):
        self.lang = 'es'
        super(InflectorEs, self).__init__()

    def _get_default_pronoun(self, person, gender='m', is_reflexive=False):
        ret = ''
        if person == '1s':
            ret = 'yo'
            if is_reflexive:
                ret += ' me'
        elif person == '2s':
            ret = 'tú'
            if is_reflexive:
                ret += ' te'
        elif person == '3s':
            ret = 'él'
            if gender == 'f':
                ret = 'ella'
            if is_reflexive:
                ret += ' se'
        elif person == '1p':
            ret = 'nosotros'
            if is_reflexive:
                ret += ' nos'
        elif person == '2p':
            ret = 'vosotros'
            if is_reflexive:
                ret += ' os'
        elif person == '3p':
            ret = 'ellos'
            if gender == 'f':
                ret = 'ellas'
            if is_reflexive:
                ret += ' se'
        return ret



# -*- coding: utf-8 -*-

class PersonEnding:
    """
    p_elem
    aka <p>
    Ending for a specific verb template, mood, tense and grammatical person
    May also have an alternate ending for an alternative spelling.
    E.g. Ending for aim:er indicatif présent 2nd Person Plural = ['ez']
    E.g. Ending for pa:yer indicatif présent 1st Person Singular = ['ie', 'ye']
    Explanation: 'ye' is an alternate spelling (je paie, je paye)
    p_elem
        Example p_elems:
            <p><i>ez</i></p>
            <p><i>eoir</i><i>oir</i></p>
            <p></p>

    person
    A grammar_defines.PERSONS value indicating which person 
    this PersonEnding is for, e.g. for aim:er, "ez" is '2p' (second person plural)
    """
    def __init__(self, p_elem, person):
        self.person = person
        self.endings = []
        for i_elem in p_elem.findall('i'):
            ending = u''
            if i_elem.text is not None:
                ending += i_elem.text
            self.endings.append(ending)

    def get_person(self):
        return self.person

    def get_ending(self):
        return self.endings[0]

    def get_alternate_ending(self):
        ret = None
        if len(self.endings) > 1:
            ret = self.endings[1]
        return ret



# -*- coding: utf-8 -*-

from . import inflector_fr
from . import inflector_es

SUPPORTED_LANGUAGES = ('fr', 'es')

class Conjugator:
    def __init__(self, lang):
        if lang == 'fr':
            self._inflector = inflector_fr.InflectorFr()
        elif lang == 'es':
            self._inflector = inflector_es.InflectorEs()
        else:
            raise InvalidLangError
    
    def conjugate(self, infinitive):
        return self._inflector.conjugate(infinitive)

    def conjugate_mood(self, infinitive, mood_name):
        return self._inflector.conjugate_mood(infinitive, mood_name)

    def conjugate_mood_tense(self, infinitive, mood_name, tense_name):
        return self._inflector.conjugate_mood_tense( 
            infinitive, mood_name, tense_name)

    def find_verb_by_infinitive(self, infinitive):
        return self._inflector.find_verb_by_infinitive(infinitive)

    def find_template(self, name):
        return self._inflector.find_template(name)

    def get_verbs_that_start_with(self, query, max_results):
        return self._inflector.get_verbs_that_start_with(query, max_results)




class ConjugatorError(Exception):
    pass

class ConjugationsParserError(Exception):
    pass

class ConjugationTemplateError(Exception):
    pass

class InvalidLangError(Exception):
    pass

class InvalidMoodError(Exception):
    pass

class InvalidTenseError(Exception):
    pass

class TemplateNotFoundError(Exception):
    pass

class VerbNotFoundError(Exception):
    pass

class VerbsParserError(Exception):
    pass



# -*- coding: utf-8 -*-

from lxml import etree
from . import string_utils
from . import exceptions

class Verb:
    def __init__(self, v_elem):
        if v_elem.tag != 'v':
            raise exceptions.VerbsParserError("Unexpected element")
        try:
            self.infinitive = u'' + v_elem.find('i').text
            self.infinitive_no_accents = \
                string_utils.strip_accents(self.infinitive)
            self.template = u'' + v_elem.find('t').text
            self.translation_en = ''
            en_node = v_elem.find('en')
            if en_node is not None:
                self.translation_en = u'' + en_node.text
            self.impersonal = False
        except AttributeError as e:
            raise exceptions.VerbsParserError(
                "Error parsing {}: {}".format(
                    etree.tostring(v_elem),
                    str(e)))



# -*- coding: utf-8 -*-

import copy

from . import inflector
from . import grammar_defines
from . import string_utils
from . import exceptions

TENSES_CONJUGATED_WITHOUT_PRONOUNS = ['infinitif-présent', 'participe-présent', 
                                      'imperatif-présent', 'participe-passé']
VERBS_CONJUGATED_WITH_ETRE = [
"aller",
"arriver",
"descendre",
"redescendre",
"entrer",
"rentrer",
"monter",
"remonter",
"mourir",
"naître",
"renaître",
"partir",
"repartir",
"passer",
"rester",
"retourner",
"sortir",
"ressortir",
"tomber",
"retomber",
"venir",
"devenir",
"parvenir",
"revenir"]

VERBS_THAT_CANNOT_BE_REFLEXIVE_OTHER_THAN_IMPERSONAL_VERBS = [
"être",
"aller",
"avoir"]

def prepend_with_que(pronoun_string):
    if string_utils.starts_with_vowel(pronoun_string):
        return "qu'" + pronoun_string
    else:
        return "que " + pronoun_string

def prepend_with_se(s):
    if string_utils.starts_with_vowel(s):
        return "s'" + s
    else:
        return "se " + s

class InflectorFr(inflector.Inflector):
    def __init__(self):
        self.lang = 'fr'
        super(InflectorFr, self).__init__()

    def get_verbs_that_start_with(self, query, max_results):
        query = query.lower()
        is_reflexive, query = self._split_reflexive(query)
        matches = self._verb_parser.get_verbs_that_start_with(query, max_results)
        if is_reflexive:
            matches = [prepend_with_se(m) 
            for m in matches if self._verb_can_be_reflexive(m)]
        return matches

    def _is_impersonal_verb(self, infinitive):
        ret = False
        verb = self.find_verb_by_infinitive(infinitive)
        template = self.find_template(verb.template)
        if len(template.moods['indicatif'].tenses['présent'].person_endings) < 6:
            ret = True
        return ret

    def _verb_can_be_reflexive(self, infinitive):
        return (not self._is_impersonal_verb(infinitive) 
            and infinitive not in 
            VERBS_THAT_CANNOT_BE_REFLEXIVE_OTHER_THAN_IMPERSONAL_VERBS) 

    def _split_reflexive(self, infinitive):
        is_reflexive = False
        if infinitive.startswith("se "):
            is_reflexive = True
            infinitive = infinitive[3:]
        elif infinitive.startswith("s'"):
            is_reflexive = True
            infinitive = infinitive[2:]
        return is_reflexive, infinitive

    def _get_pronoun_suffix(self, person, gender='m'):
        return '-' + self._get_default_pronoun(person, gender).replace('tu', 'toi')

    def _get_default_pronoun(self, person, gender='m', is_reflexive=False):
        ret = ''
        if person == '1s':
            ret = 'je'
            if is_reflexive:
                ret += ' me'
        elif person == '2s':
            ret = 'tu'
            if is_reflexive:
                ret += ' te'
        elif person == '3s':
            ret = 'il'
            if gender == 'f':
                ret = 'elle'
            if is_reflexive:
                ret += ' se'
        elif person == '1p':
            ret = 'nous'
            if is_reflexive:
                ret += ' nous'
        elif person == '2p':
            ret = 'vous'
            if is_reflexive:
                ret += ' vous'
        elif person == '3p':
            ret = 'ils'
            if gender == 'f':
                ret = 'elles'
            if is_reflexive:
                ret += ' se'
        return ret

    def _conjugate_simple_mood_tense(self, verb_stem, mood_name, 
                                     tense_template, is_reflexive=False):
        ret = []
        if tense_template.name in TENSES_CONJUGATED_WITHOUT_PRONOUNS:
            for person_ending in tense_template.person_endings:
                conj = ''
                if is_reflexive and tense_template.name == 'participe-passé':
                    conj += 'étant '
                conj += verb_stem + person_ending.get_ending()
                if is_reflexive:
                    if mood_name != 'imperatif':
                        conj = prepend_with_se(conj)
                    else:
                        conj += self._get_pronoun_suffix(person_ending.get_person())
                ret.append(conj)
        else:
            for person_ending in tense_template.person_endings:
                pronoun = self._get_default_pronoun(
                    person_ending.get_person(), is_reflexive=is_reflexive)
                ending = person_ending.get_ending()

                conjugation = ''
                conjugated_verb = verb_stem + ending
                if pronoun[-1] == "e" and string_utils.starts_with_vowel(conjugated_verb):
                    conjugation += pronoun[:-1] + "'"
                else:
                    conjugation += pronoun + " "
                conjugation += conjugated_verb

                if mood_name == 'subjonctif':
                    conjugation = prepend_with_que(conjugation)
                ret.append(conjugation)
        return ret

    def _get_compound_conjugations_hv_map(self):
        return {
            'indicatif': {
                'passé-composé': 'présent',
                'plus-que-parfait': 'imparfait',
                'futur-antérieur': 'futur-simple',
                'passé-antérieur': 'passé-simple'
            },
            'subjonctif': {
                'passé': 'présent',
                'plus-que-parfait': 'imparfait'
            },
            'conditionnel': {
                'passé': 'présent'
            },
            'imperatif': {
                'imperatif-passé': 'imperatif-présent'
            }
        }

    def _conjugate_compound(self, co, mood_name, hv_tense_name):
        """Conjugate a compound tense
        Args:
            co: ConjugationObjects for the verb being conjugated
            mood_name: mood verb is being conjugated in
            hv_tense_name: tense_name for conjugating helping verb
        """
        ret = []
        if (co.is_reflexive and mood_name == 'imperatif' 
            and hv_tense_name == 'imperatif-présent'):
            return ret
        persons = [pe.person for pe in 
            co.template.moods[mood_name].tenses[hv_tense_name].person_endings]
        helping_verb = 'avoir'
        if (co.verb.infinitive in VERBS_CONJUGATED_WITH_ETRE
            or co.is_reflexive):
            helping_verb = 'être'
        hvco = self._get_conj_obs(helping_verb)
        hvtense_template = copy.deepcopy(
            hvco.template.moods[mood_name].tenses[hv_tense_name])
        hvperson_endings = []
        for pe in hvtense_template.person_endings:
            if pe.person in persons:
                hvperson_endings.append(pe)
        hvtense_template.person_endings = hvperson_endings
        hvconj = self._conjugate_simple_mood_tense(
            hvco.verb_stem, 
            'indicatif', 
            hvtense_template,
            co.is_reflexive)
        participle = self._conjugate_simple_mood_tense(
            co.verb_stem, 
            'participe', 
            co.template.moods['participe'].tenses['participe-passé'])
        if helping_verb == 'avoir':
            for hv in hvconj:
                p = participle[0]
                ret.append(hv + ' ' + p)
        else:
            for i, hv in enumerate(hvconj):
                participle_inflection = \
                    self._get_default_participle_inflection_for_person(
                        persons[i])
                p = participle[
                    grammar_defines.PARTICIPLE_INFLECTIONS.index(
                        participle_inflection)]
                ret.append(hv + ' ' + p)
        if mood_name == 'subjonctif':
            ret = [prepend_with_que(i) for i in ret]
        return ret



# -*- coding: utf-8 -*-

import unicodedata

from verbecc import grammar_defines


def strip_accents(s):
    return ''.join(c for c in unicodedata.normalize('NFD', s)
                   if unicodedata.category(c) != 'Mn')

def starts_with_vowel(s):
    if len(s) == 0:
        return False
    return strip_accents(s)[0] in grammar_defines.VOWELS

def unicodefix(s):
    # Fix Python 2.x.
    try:
        return s.decode('utf-8')
    except AttributeError:
        return s



# -*- coding: utf-8 -*-

from .tense_template import TenseTemplate

class Mood():
    def __init__(self, mood_elem):
        self.name = mood_elem.tag
        self.tenses = {}
        for tense_elem in list(mood_elem):
            self.tenses[tense_elem.tag] = TenseTemplate(tense_elem)



# -*- coding: utf-8 -*-

from lxml import etree

from .mood import Mood
from . import exceptions

class ConjugationTemplate:
    def __init__(self, template_elem):
        if template_elem.tag != 'template':
            raise exceptions.ConjugationTemplateError(
                "Unexpected element")
        try:
            self.name = u'' + template_elem.get('name')
            self.moods = {}
            for mood_elem in list(template_elem):
                mood = Mood(mood_elem)
                self.moods[mood_elem.tag] = mood
                    
        except AttributeError as e:
            raise exceptions.ConjugationTemplateError(
                "Error parsing {}: {}".format(
                    etree.tostring(template_elem),
                    str(e)))



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod

from . import grammar_defines
from . import string_utils
from . import exceptions
from . import parse_verbs
from . import parse_conjugations

class Inflector(ABC):
    def __init__(self):
        self._verb_parser = parse_verbs.VerbsParser(self.lang)
        self._conj_parser = parse_conjugations.ConjugationsParser(self.lang)

    def conjugate(self, infinitive):
        co = self._get_conj_obs(infinitive)
        moods = {}
        for mood in co.template.moods:
            moods[mood] = self._conjugate_mood(co, mood)
        return {'verb': {'infinitive': co.verb.infinitive, 
                         'template': co.verb.template,
                         'translation_en': co.verb.translation_en,
                         'stem': co.verb_stem}, 
                'moods': moods}

    def conjugate_mood(self, infinitive, mood_name):
        co = self._get_conj_obs(infinitive)
        return self._conjugate_mood(co, mood_name)

    def conjugate_mood_tense(self, infinitive, mood_name, tense_name):
        co = self._get_conj_obs(infinitive)
        return self._conjugate_mood_tense(co, mood_name, tense_name)

    def find_verb_by_infinitive(self, infinitive):
        return self._verb_parser.find_verb_by_infinitive(infinitive)

    def find_template(self, name):
        return self._conj_parser.find_template(name)

    def get_verbs_that_start_with(self, query, max_results):
        query = query.lower()
        matches = self._verb_parser.get_verbs_that_start_with(query, max_results)
        return matches

    def _get_verb_stem(self, infinitive, template_name):
        template_beg, template_ending = template_name.split(u':')
        if not infinitive.endswith(template_ending):
            raise exceptions.ConjugatorError(
                "Template {} ending doesn't "
                "match infinitive {}"
                .format(template_name, infinitive))
        return infinitive[:len(infinitive) - len(template_ending)]

    def _is_impersonal_verb(self, infinitive):
        return False

    def _verb_can_be_reflexive(self, infinitive):
        return not self._is_impersonal_verb(infinitive)

    def _split_reflexive(self, infinitive):
        return (False, infinitive)

    class ConjugationObjects:
        def __init__(self, infinitive, verb, template, verb_stem, is_reflexive):
            self.infinitive = infinitive
            self.verb = verb
            self.template = template
            self.verb_stem = verb_stem
            self.is_reflexive = is_reflexive

    def _get_conj_obs(self, infinitive):
        infinitive = infinitive.lower()
        is_reflexive, infinitive = self._split_reflexive(infinitive)
        if is_reflexive and not self._verb_can_be_reflexive(infinitive):
            raise exceptions.VerbNotFoundError("Verb cannot be reflexive")
        verb = self.find_verb_by_infinitive(infinitive)
        template = self.find_template(verb.template)
        verb_stem = self._get_verb_stem(verb.infinitive, template.name)
        return Inflector.ConjugationObjects(
            infinitive, verb, template, verb_stem, is_reflexive)     

    def _conjugate_mood(self, co, mood_name):
        if mood_name not in co.template.moods:
            raise exceptions.InvalidMoodError
        ret = {}
        ret.update(self._get_simple_conjugations_for_mood(co, mood_name))
        ret.update(self._get_compound_conjugations_for_mood(co, mood_name))
        return ret

    def _get_simple_conjugations_for_mood(self, co, mood_name):
        ret = {}
        mood = co.template.moods[mood_name]
        for tense_name in mood.tenses:
            ret[tense_name] = self._conjugate_mood_tense(co, mood_name, tense_name)
        return ret

    def _get_compound_conjugations_for_mood(self, co, mood_name):
        ret = {}
        comp_conj_map = self._get_compound_conjugations_hv_map()
        if mood_name in comp_conj_map:
            for tense_name in comp_conj_map[mood_name]:
                ret[tense_name] = self._conjugate_mood_tense(co, mood_name, tense_name)
        return ret

    def _conjugate_mood_tense(self, co, mood_name, tense_name):
        comp_conj_map = self._get_compound_conjugations_hv_map()
        if mood_name in comp_conj_map and tense_name in comp_conj_map[mood_name]:
            hv_tense_name = comp_conj_map[mood_name][tense_name]
            return self._conjugate_compound(co, mood_name, hv_tense_name)
        else:
            mood = co.template.moods[mood_name]
            if tense_name not in mood.tenses:
                raise exceptions.InvalidTenseError
            tense_template = mood.tenses[tense_name]
            return self._conjugate_simple_mood_tense(
                co.verb_stem, mood_name, tense_template,
                co.is_reflexive)

    def _get_compound_conjugations_hv_map(self):
        """"Returns a map of the tense of the helping verb for each compound mood and tense"""
        return {}

    def _conjugate_compound(self, co, mood_name, hv_tense_name):
        """Conjugate a compound tense
        Args:
            co: ConjugationObjects for the verb being conjugated
            mood_name: mood verb is being conjugated in
            hv_tense_name: tense_name for conjugating helping verb
        """
        return {}

    def _get_default_participle_inflection_for_person(self, person):
        if person[1] == 's':
            return 'ms'
        else:
            return 'mp'

    def _get_default_pronoun(self, person, gender='m', is_reflexive=False):
        return ''

    def _conjugate_simple_mood_tense(self, verb_stem, mood_name, 
                                     tense_template, is_reflexive=False):
        ret = []
        for person_ending in tense_template.person_endings:
            pronoun = self._get_default_pronoun(
                person_ending.get_person(), is_reflexive=is_reflexive)
            ending = person_ending.get_ending()

            conjugation = ''
            conjugated_verb = verb_stem + ending
            if pronoun[-1] == "e" and string_utils.starts_with_vowel(conjugated_verb):
                conjugation += pronoun[:-1] + "'"
            else:
                conjugation += pronoun + " "
            conjugation += conjugated_verb

            if mood_name == 'subjonctif':
                conjugation = prepend_with_que(conjugation)
            ret.append(conjugation)
        return ret



#!/usr/bin/env python
from setuptools import setup, find_packages
from verbecc import __version__

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    "lxml>=4.1.1",
    "mock>=2.0.0,<3.0.0",
    "pytest>=3.0.6,<4.0.0",
    "pytest-cov>=2.4.0,<3.0.0",
    "pylama>=7.4.1,<8.0.0"
]

setup(
    name='verbecc',
    version=__version__,
    description='Verbes, complètement conjugués - conjugueur français | Verbs, completely conjugated - French conjugator',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Brett Tolbert',
    author_email='brett.tolbert@gmail.com',
    url='https://github.com/bretttolbert/verbecc',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=install_requires,
    test_suite="tests",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Natural Language :: French",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Education",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic"
    ],
)


# -*- coding: utf-8 -*-

VOWELS = ('a', 'e', 'i', 'o', 'u')

# This refers to grammatical person ('usted' is 3s despite being semantically 2p)
PERSONS = ('1s', '2s', '3s', '1p', '2p', '3p')

IMPERATIVE_PRESENT_PERSONS = ('2s', '1p', '2p')

PARTICIPLE_INFLECTIONS = ('ms', 'mp', 'fs', 'fp')
